<?php

namespace Drupal\cryptolog;

use Drupal\Core\Cache\ApcuBackendFactory;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\ChainedFastBackendFactory;
use Drupal\Core\Site\Settings;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Cryptolog middleware.
 */
class CryptologMiddleware implements HttpKernelInterface {

  /**
   * The cache bin.
   */
  const BIN = 'bootstrap';

  /**
   * The cache identifier.
   */
  const KEY = 'cryptolog';

  /**
   * The cache time to live in seconds.
   */
  const TTL = 86400;

  /**
   * The cache backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $backend;

  /**
   * The cached salt and day of year.
   *
   * @var string[]
   */
  protected $cache;

  /**
   * The original client IP address.
   *
   * @var string|null
   */
  protected $clientIp;

  /**
   * The HTTP host if it was set by Cryptolog.
   *
   * @var string|null
   */
  protected $httpHost;

  /**
   * The decorated kernel.
   *
   * @var \Symfony\Component\HttpKernel\HttpKernelInterface
   */
  protected $httpKernel;

  /**
   * Whether or not this request is from a trusted reverse proxy.
   *
   * @var bool|null
   */
  protected $isFromTrustedProxy;

  /**
   * Whether or not the request is secure, if set by Cryptolog.
   *
   * @var bool|null
   */
  protected $isSecure;

  /**
   * Constructs the Cryptolog middleware.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   The decorated kernel.
   * @param \Drupal\Core\Site\Settings $settings
   *   The settings object.
   * @param \Drupal\Core\Cache\ChainedFastBackendFactory $chained_fast
   *   Chained fast cache backend factory.
   * @param \Drupal\Core\Cache\ApcuBackendFactory $apcu
   *   APCu cache backend factory.
   * @param \Drupal\memcache\MemcacheBackendFactory|null $memcache
   *   (optional) Memcache cache backend factory.
   * @param \Drupal\memcache_storage\MemcachedBackendFactory|null $memcache_storage
   *   (optional) Memcache Storage cache backend factory.
   * @param \Drupal\redis\Cache\CacheBackendFactory|null $redis
   *   (optional) Redis cache backend factory.
   */
  public function __construct(HttpKernelInterface $http_kernel, Settings $settings, ChainedFastBackendFactory $chained_fast, ApcuBackendFactory $apcu, $memcache = NULL, $memcache_storage = NULL, $redis = NULL) {
    $this->httpKernel = $http_kernel;
    $this->setBackend($settings, $chained_fast, $apcu, $memcache, $memcache_storage, $redis);
    $this->cache = $this->getCache();
    $day = gmdate('zY');
    if (!$this->cache || $this->cache['day'] !== $day) {
      $this->cache = ['day' => $day, 'salt' => random_bytes(32)];
      $this->setCache($this->cache);
    }
  }

  /**
   * Returns the backend cache bin for diagnostic purposes.
   */
  public function getBackend(): CacheBackendInterface {
    return $this->backend;
  }

  /**
   * Fetches salt and day of year from the cache.
   *
   * @return string[]
   *   The array of cached data, or an empty array.
   */
  protected function getCache(): array {
    $cache = $this->backend->get(self::KEY);
    return $cache ? $cache->data : [];
  }

  /**
   * Returns the original client IP address for diagnostic purposes.
   */
  public function getClientIp(): ?string {
    return $this->clientIp;
  }

  /**
   * Returns the HTTP host if it was set by Cryptolog.
   */
  public function getHttpHost(): ?string {
    return $this->httpHost;
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-ignore-next-line Not yet using MAIN_REQUEST.
   */
  public function handle(Request $request, $type = self::MASTER_REQUEST, $catch = TRUE): Response {
    // @phpstan-ignore-next-line Not yet using MAIN_REQUEST.
    if ($type === self::MASTER_REQUEST) {
      $this->setClientIp($request);
    }
    return $this->httpKernel->handle($request, $type, $catch);
  }

  /**
   * Returns whether or not the request is secure, if set by Cryptolog.
   */
  public function isSecure(): ?bool {
    return $this->isSecure;
  }

  /**
   * Returns whether or not the request is from a trusted reverse proxy.
   */
  public function isFromTrustedProxy(): ?bool {
    return $this->isFromTrustedProxy;
  }

  /**
   * Sets the backend cache bin.
   *
   * @param \Drupal\Core\Site\Settings $settings
   *   The settings object.
   * @param \Drupal\Core\Cache\ChainedFastBackendFactory $chained_fast
   *   Chained fast cache backend factory.
   * @param \Drupal\Core\Cache\ApcuBackendFactory $apcu
   *   APCu cache backend factory.
   * @param \Drupal\memcache\MemcacheBackendFactory|null $memcache
   *   (optional) Memcache cache backend factory.
   * @param \Drupal\memcache_storage\MemcachedBackendFactory|null $memcache_storage
   *   (optional) Memcache Storage cache backend factory.
   * @param \Drupal\redis\Cache\CacheBackendFactory|null $redis
   *   (optional) Redis cache backend factory.
   */
  protected function setBackend(Settings $settings, ChainedFastBackendFactory $chained_fast, ApcuBackendFactory $apcu, $memcache = NULL, $memcache_storage = NULL, $redis = NULL): void {
    // Fall back to chained fast cache backend if APCu extension is disabled.
    if (!extension_loaded('apcu')) {
      $apcu = NULL;
    }
    $backend = $memcache ?: $memcache_storage ?: $redis ?: $apcu ?: $chained_fast;
    // Use chained fast cache backend if default cache backend is memory-based.
    $cache_settings = $settings->get('cache');
    if (isset($cache_settings['default'])) {
      switch ($cache_settings['default']) {
        case 'cache.backend.memcache':
        case 'cache.backend.memcache_storage':
        case 'cache.backend.redis':
          $backend = $chained_fast;
      }
    }
    try {
      $this->backend = $backend->get(self::BIN);
    }
    catch (\Exception $e) {
      // Redis module is installed but a Redis client is not available.
      $backend = $apcu ?: $chained_fast;
      $this->backend = $backend->get(self::BIN);
    }
  }

  /**
   * Stores salt and day of year in the cache.
   *
   * @param string[] $value
   *   The array of data to cache.
   */
  protected function setCache(array $value): void {
    $this->backend->set(self::KEY, $value, self::TTL + time());
  }

  /**
   * Sets the client IP address.
   */
  protected function setClientIp(Request $request): void {
    // Once the client IP address is modified, the request will no longer be
    // recognized as coming from a trusted reverse proxy, and therefore the
    // scheme and HTTP host can no longer be dynamically extracted from the HTTP
    // headers (if that's where they were coming from). Save the scheme and HTTP
    // host so they can be restored to the request if necessary.
    if ($isFromTrustedProxy = $request->isFromTrustedProxy()) {
      $isSecure = $request->isSecure();
      $httpHost = $request->getHttpHost();
    }

    $this->clientIp = $request->getClientIp();
    // Generate a keyed hash using sodium extension, if available.
    $hmac = function_exists('sodium_crypto_generichash') ? sodium_crypto_generichash($this->clientIp, $this->cache['salt'], 16) : hash_hmac('md5', $this->clientIp, $this->cache['salt'], TRUE);
    // Convert to IPv6 address notation.
    $request->server->set('REMOTE_ADDR', inet_ntop($hmac));

    // For reverse proxy requests, restore the scheme and HTTP host if they do
    // not match their previous values. Record the changes we're making for
    // diagnostic purposes.
    if ($isFromTrustedProxy) {
      if ($request->isSecure() !== $isSecure) {
        $request->server->set('HTTPS', $isSecure ? 'on' : 'off');
        $this->isSecure = $isSecure;
      }
      if ($request->getHttpHost() !== $httpHost) {
        $request->headers->set('HOST', $httpHost);
        $this->httpHost = $httpHost;
      }
    }
    $this->isFromTrustedProxy = $isFromTrustedProxy;
  }

}
